program InSync;

{*************************************************************************
*                                                                        *
*                    Directory Tree File Synchroniser                    *
*                  Copyright (C) 1997-98, John Zaitseff                  *
*                                                                        *
*************************************************************************}

{ Author:    John Zaitseff <J.Zaitseff@unsw.edu.au>
  Date:      11th April, 1998.
  Version:   1.2

  This file contains the main source code for the Directory Tree File
  Synchroniser (INSYNC).  Please see INSYNC.TXT for an explanation of
  the command line interface.

  This program is free software, and may be distributed and/or modified
  under the terms of the GNU General Public License as published by the
  Free Software Foundation; either Version 2 of the License or any later
  version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program (in the file LICENSE.TXT); if not, write to:
            Free Software Foundation, Inc.,
            675 Mass Ave, Cambridge, MA 02139, USA.

  Please feel free to send any comments, suggestions, corrections and
  enhancements to:

  Postal:   John Zaitseff,
            4 McCabe Close,
            Menai, NSW, 2234,
            Australia.
  Email:    J.Zaitseff@unsw.edu.au
  Web:      http://biosys12.ee.unsw.edu.au/personal/software/utilities/

}


{$IFNDEF WIN32}
Error: 'Compile under Borland Delphi for Win32 only (v2 or later).'
{$ENDIF}

{$IFNDEF OPTDEBUG}
{$UNDEF DEBUG}
{$ENDIF}

{$APPTYPE CONSOLE}
{$DESCRIPTION 'Directory Tree File Synchroniser'}
{$RESOURCE *.res}

{$A+,B-,H+,J-,O+,P-,T+,U+,V+,W-,X+,Z1}
{$IFDEF DEBUG}
    {$D+,I+,L+,Q+,R+,Y+}
    {$WARNINGS ON}
    {$HINTS ON}
{$ELSE}
    {$D-,I+,L-,Q-,R-,Y-}
    {$WARNINGS ON}
    {$HINTS ON}
{$ENDIF}


uses
  Windows, SysUtils,
  ParamLn, VerRes, DirEx, MiscStr;


type
  { Operating mode: Master-slave, Absolute Master-slave, Peer-peer }
  TOpMode    = (MasterSlave, AbsMasterSlave, PeerPeer);
  TOpModeStr = array [TOpMode] of string;


const
  { Miscellaneous constants }
  DefTmpPrefix       = 'IS';     { 2-letter prefix for temporary filenames }

  { String resource constants }
  {$INCLUDE InSync.inc}


{***********************************************************************}
{ Display the standard program heading }
procedure ShowHeading;

begin
  writeln(Format(s_Heading, [AppVerRes.FileVersion]))
end;


{***********************************************************************}
{ Display command line usage, then terminate }
procedure ShowUsage;

begin
  writeln(s_Usage);
  halt(0)
end;


{***********************************************************************}
{ Display an error message, with an optional hint, and terminate with
  error code 1 }
procedure ShowError (const Msg : string; ShowHint : boolean);

begin
  writeln(s_Error, Msg);
  if ShowHint then
    writeln(s_ErrHint);

  halt(1)
end;


{***********************************************************************}
{ Display an error message, together with the last system diagnostic, and
  terminate with error code 1 }
procedure ShowSysError (const Msg : string);

begin
  writeln(s_Error, Msg);
  writeln(s_SysError, SysErrorMessage(GetLastError));

  halt(1)
end;


{***********************************************************************}
{ Expand the directory name passed into a fully-qualified directory name.
  This routine adds a trailing backslash if not present.  A null string
  is returned if an error is encountered (primarily if an improper UNC
  path is specified; use GetLastError to determine the exact error
  message) }
function ExpandDirNameEx (const DirName : string) : string;

begin
  Result := DirName;
  if Result = '' then
    Result := '.\';

  Result := ExpandFileName(Result);
  if Result = '' then
    exit;

  if not (Result[length(Result)] in [':', '\']) then
    Result := Result + '\'
end;


{***********************************************************************}
{ Strip off trailing backslashes from the passed string, if appropriate.
  If the string is a drive name and backslash, do not strip the backslash }
function StripBackslash (const S : string) : string;

var
  FileDriveLen : integer;

begin
  FileDriveLen := length(ExtractFileDrive(S));
  Result := S;
  while (Result <> '') and
        (length(Result) > FileDriveLen + 1) and
        (Result[length(Result)] = '\') do
    delete(Result, length(Result), 1)
end;


{***********************************************************************}
{ Search the directory entry list DirEntryList for an entry that has
  a filename of Filename (case insensitive match), and return the
  position in the list if found, or -1 if not }
function DirEntrySearchFN (DirEntryList : TDirEntryList;
                           const Filename : string) : integer;

var
  I : integer;

begin
  Result := -1;
  for I := 0 to DirEntryList.Count - 1 do
    if CompareText(DirEntryList[I].Name, Filename) = 0 then
      begin
        Result := I;
        exit
      end
end;


{***********************************************************************}
{ Compare two DOS date/times and return a number < 0 if D1 is earlier
  than D2, 0 if D1 and D2 are the same, or a number > 0 if D1 is later
  than D2 }
function CompareDOSTimes (D1, D2 : integer) : integer;

var
  TD1, TD2 : TDateTime;

begin
  TD1 := FileDateToDateTime(D1);
  TD2 := FileDateToDateTime(D2);
  if TD1 = TD2 then
    Result := 0
  else if TD1 < TD2 then
    Result := -1
  else
    Result := 1
end;


{***********************************************************************}
{ Return a unique filename, with a directory path of Path (which must
  end in a ':' or '\') and having a prefix of Prefix (up to the first
  two characters).  This function checks that the file does not exist,
  but does not create anything.  If Path is blank, the TEMP directory
  is used (as returned by GetTempPath).  Prefix must only contain
  alphanumeric and/or underscore characters, or else a default prefix
  will be used.  The file extension is specified by the constant DefExt
  below.  Note that if Path is invalid, a blank string will be returned
  instead of a filename }
function UniqueFilename (Path, Prefix : string) : string;

const
  DefPrefix = '';       { Default prefix if the supplied one is invalid }
  DefExt    = '.$$$';   { Extension to use for temporary filename }

var
  N, ErrNum   : integer;
  H, M, S, MS : word;
  SearchRec   : TSearchRec;

begin
  if Path = '' then
    begin
      SetLength(Path, MAX_PATH);
      N := GetTempPath(length(Path), PChar(Path));
      if N <= 0 then
        Path := ''
      else
        SetLength(Path, N)
    end;

  Path := ExpandDirNameEx(Path);
  if Path = '' then
    begin
      Result := '';
      exit
    end;

  if length(Prefix) > 2 then
    SetLength(Prefix, 2);
  if not IsValidIdent(Prefix) then
    Prefix := DefPrefix;

  Path := Path + Prefix;

  { Use current time for unique part of string }
  DecodeTime(Time, H, M, S, MS);
  N := H * 36000 + M * 600 + S * 10 + MS div 100;

  repeat
    Result := Path + IntToHex(N, 6) + DefExt;

    ErrNum := FindFirst(Result, faAnyFile, SearchRec);
    FindClose(SearchRec);
    if ErrNum = 0 then
      { File of the same name exists; generate a new filename }
      inc(N)
  until ErrNum <> 0
end;


{***********************************************************************}
{ Process the command line, checking for any errors, etc.  Sets DirName1
  to the first directory specified, DirName2 to the second directory and
  OpMode to the operating mode:  MasterSlave, AbsMasterSlave, PeerPeer.
  Also sets the ForceAction, BeQuieter, FirstPhaseOnly and Simulating
  flags, as appropriate. }
procedure ProcessCmdLine (var DirName1, DirName2 : string;
                          var OpMode : TOpMode;
                          var ForceActions, BeQuieter,
                              FirstPhaseOnly, Simulating : boolean);

var
  OpModeSet : boolean;

  { Internal procedure }
  procedure SetOpMode (I : integer; NewMode : TOpMode);

  begin
    if OpModeSet and (OpMode <> NewMode) then
      ShowError(Format(s_ErrSettingOpMode, [UpperCase(Par.Str(I))]), False)
    else
      begin
        OpMode := NewMode;
        OpModeSet := True
      end
  end;

var
  I : integer;

begin { ProcessCmdLine }
  if (Par.Option('?', 0) <> 0) or (Par.Count < 1) then
    ShowUsage;

  DirName1 := '';
  DirName2 := '';

  { Set default values for parameters }
  ForceActions := False;
  BeQuieter := False;
  FirstPhaseOnly := False;
  Simulating := False;

  OpMode := PeerPeer;
  OpModeSet := False;

  { Process command line }
  for I := 1 to Par.Count do
    begin
      if not Par.IsOption(I) then
        begin
          { Not a command line option: must be a directory name }
          if DirName1 = '' then
            DirName1 := Par.Str(I)
          else if DirName2 = '' then
            DirName2 := Par.Str(I)
          else
            ShowError(s_ErrTooManyParams, True)
        end
      else
        begin
          { Command line option }
          if Par.IsTheOption(I, 'PP') then          { /PP: Peer-peer }
            SetOpMode(I, PeerPeer)
          else if Par.IsTheOption(I, 'P') then      { /P:  Peer-peer }
            SetOpMode(I, PeerPeer)

          else if Par.IsTheOption(I, 'MS') then     { /MS: Master-slave }
            SetOpMode(I, MasterSlave)
          else if Par.IsTheOption(I, 'M') then      { /M:  Master-slave }
            SetOpMode(I, MasterSlave)

          else if Par.IsTheOption(I, 'ABS') then    { /ABS: Absolute master-slave }
            SetOpMode(I, AbsMasterSlave)
          else if Par.IsTheOption(I, 'A') then      { /A:   Absolute master-slave }
            SetOpMode(I, AbsMasterSlave)

          else if Par.IsTheOption(I, 'F') then      { /F:  Force actions }
            ForceActions := True
          else if Par.IsTheOption(I, 'Q1') then     { /Q1: Slightly quieter mode }
            BeQuieter := True
          else if Par.IsTheOption(I, '1') then      { /1:  First phase only }
            FirstPhaseOnly := True
          else if Par.IsTheOption(I, 'X') then      { /X:  Simulate only }
            Simulating := True

          else
            ShowError(Format(s_ErrCmdLineOption, [UpperCase(Par.Str(I))]), True)
        end
    end;

  if (DirName1 = '') or (DirName2 = '') then
    ShowError(s_ErrNotEnoughDirs, True)
end;


{***********************************************************************}
{ Validate the passed directory name:  expand the directory name into a
  fully-qualified path name, and check to see that the directory exists }
procedure ValidateDir (var DirName : string);

var
  DN        : string;
  ErrNum    : integer;
  SearchRec : TSearchRec;

begin
  DN := ExpandDirNameEx(DirName);
  if DN = '' then
    ShowSysError(Format(s_ErrInvalidDir, [DirName]));

  DirName := StripBackslash(DN);
  if (DirName <> '') and (DirName[length(DirName)] <> '\') then
    DirName := DirName + '\';

  { Try to read in a directory entry from the specified directory }
  ErrNum := FindFirst(DirName + '*.*', faAnyFile, SearchRec);
  FindClose(SearchRec);
  if ErrNum <> 0 then
    ShowSysError(Format(s_ErrInvalidDir, [StripBackslash(DirName)]));
end;


{***********************************************************************}
{ Show program sub-heading: directory names and operating mode }
procedure ShowSubHeading (const DirName1, DirName2 : string;
                          const OpMode : TOpMode;
                          const ForceActions, BeQuieter,
                                FirstPhaseOnly, Simulating : boolean);

begin
  writeln(Format(s_Synchronising, [sa_OpMode[OpMode]]));

  if OpMode = PeerPeer then
    writeln(Format(s_DirNamesPP, [StripBackslash(DirName1), StripBackslash(DirName2)]))
  else
    writeln(Format(s_DirNamesMS, [StripBackslash(DirName1), StripBackslash(DirName2)]));

  if ForceActions then
    writeln(s_ForceActions);
  if FirstPhaseOnly then
    writeln(s_FirstPhaseOnly);
  if BeQuieter then
    writeln(s_QuieterMode);
  if Simulating then
    writeln(s_Simulating);
  writeln
end;


{***********************************************************************}
{ Check for identical directory names and abort the program if this is
  the case.  This is not entirely fool-proof }
procedure CheckForIdenticalDirs (const DirName1, DirName2 : string);

begin
  if CompareText(DirName1, DirName2) = 0 then
    ShowError(s_ErrIdenticalDirs, False)
end;


{***********************************************************************}
{ Search the specified directory recursively and store all of the files
  found.  The DirList object is used for this.  Note that DirList is
  assumed to already have been constructed via Create }
procedure SearchDir (const DirName : string; DirList : TDirEntryList);

var
  Count : integer;

begin
  write(Format(s_SearchingDir, [StripBackslash(DirName)]));

  Count := AddDirEntries(DirList, DirName + '*.*', True, AllFiles);
  if Count < 0 then
    begin
      writeln;
      ShowSysError(Format(s_ErrDirNotAdded, [StripBackslash(DirName)]))
    end;

  writeln(Format(s_NumEntries, [Count]))
end;


{***********************************************************************}
{ Utility function of ProcessDirEntry.  Set the attributes of FName
  to FAttr; abort with an error if this could not be done }
procedure DoSetAttributes (const FName : string; const FAttr : integer);

begin
  if FileSetAttr(FName, FAttr) <> 0 then
    begin
      writeln;
      ShowSysError(Format(s_ErrChangeAttr, [FName]))
    end
end;


{***********************************************************************}
{ Utility function of ProcessDirEntry.  Copy file FSrc to FDest, unless
  FDest is read-only and ForceActions is False.  Copying the files also
  resets the Archive attributes.  FNameSfx is the filename suffix, as
  per ProcessDirEntry.  vids_MsgStr is the message string printed out
  to the console; vids_Dest specifies the base of the destination message
  strings (ie, either ids_OpModeDest1 or ids_OpModeDest2).  This function
  returns True if something was actually done, or aborts if an error
  occurred }
function DoCopyFile (const FSrc, FDest : TDirEntry; const FNameSfx : string;
                     const OpMode : TOpMode;
                     const ForceActions, Simulating : boolean;
                     const vs_MsgStr : string;
                     const vs_Dest : TOpModeStr) : boolean;

var
  Attr : integer;

begin
  Result := False;

  write(Format(vs_MsgStr, [FNameSfx, vs_Dest[OpMode]]));

  if not Simulating then
    begin
      if (FDest.Attr and faReadOnly) <> 0 then
        begin
          { Destination is read-only }
          if ForceActions then
            DoSetAttributes(FDest.Name, FDest.Attr and not faReadOnly)
          else
            begin
              { Skip copying the file }
              writeln(s_SkippedRO);
              exit
            end
        end;

      { Perform the actual file copying }
      if not CopyFile(PChar(FSrc.Name), PChar(FDest.Name), False) then
        begin
          writeln;
          ShowSysError(Format(s_ErrCopyFile, [FSrc.Name]))
        end;

      { Set attributes of both files, resetting the Archive attribute }
      Attr := FSrc.Attr and not faArchive;
      DoSetAttributes(FSrc.Name, Attr);
      DoSetAttributes(FDest.Name, Attr);

      Result := True
    end;

  writeln(s_Done)
end;


{***********************************************************************}
{ Utility function of ProcessDirEntry.  Check if the filename portions
  of FNOld and FNNew are case-identical.  If not (and Simulating is False),
  rename FNOld to become FNNew.  FNameSfx is the suffix part of the filenames,
  as per ProcessDirEntry.  vids_MsgStr is the message string printed
  to the console; vids_ErrStr is the error message printed if the rename was
  not successful.  vids_Dest specifies the base of the destination message
  strings.  This function returns True if something was actually
  done, or aborts if an error occurred.  If BeQuieter is True, nothing is
  printed if the function is successful }
function DoRename (const FNOld, FNNew, FNameSfx : string;
                   const OpMode : TOpMode;
                   const BeQuieter, Simulating : boolean;
                   const vs_MsgStr, vs_ErrStr : string;
                   const vs_Dest : TOpModeStr) : boolean;

var
  TempFN,
  FPath,
  FNNewFN : string;

begin
  Result := False;

  FNNewFN := ExtractFileName(FNNew);
  if ExtractFileName(FNOld) <> FNNewFN then
    begin
      { Filenames are NOT the same:  rename FNOld to become FNNew }
      if not BeQuieter then
        write(Format(vs_MsgStr, [FNameSfx, vs_Dest[OpMode]]));

      if not Simulating then
        begin
          FPath := ExtractFilePath(FNOld);

          { Generate a unique filename for temporary purposes }
          TempFN := UniqueFileName(FPath, DefTmpPrefix);
          if TempFN = '' then
            begin
              { Could not generate temporary filename }
              writeln;
              ShowSysError(Format(vs_ErrStr, [FNOld]))
            end
          else
            begin
              { Rename old filename to the temporary filename first }
              if not RenameFile(FNOld, TempFN) then
                begin
                  writeln;
                  ShowSysError(Format(vs_ErrStr, [FNOld]))
                end;

              { Rename temporary name to final name FNNewFN }
              if not RenameFile(TempFN, FPath + FNNewFN) then
                begin
                  { Error occurred: try not to leave FNOld as the
                    temporary filename! }
                  RenameFile(TempFN, FNOld);
                  writeln;
                  ShowSysError(Format(vs_ErrStr, [FNOld]))
                end;

              Result := True
            end
        end;

      if not BeQuieter then
        writeln(s_Done)
    end
end;


{***********************************************************************}
{ Sub-function of ProcessDirEntry, called when a directory exists in
  both DL1 and DL2 directory entry lists.  This function checks to see
  that the second entry IS a directory (the first one is known to be,
  from the logic in ProcessDirEntry), then makes sure that the directory
  names are the same, case sensitively; if not, rename the second
  directory entry.  This function does NOT check that the date/time
  stamps are the same, since Win32/DOS does not honour FileSetDate on
  directories.  True is returned if something was actually done }
function ProcessExistingDirectory (const DL1I, DL2J : TDirEntry;
                                   const FNameSfx : string;
                                   const OpMode : TOpMode;
                                   const BeQuieter, Simulating : boolean;
                                   const vs_Dest : TOpModeStr) : boolean;

begin
  { Check that DL2J is a directory }
  if not ((DL2J.Attr and faDirectory) <> 0) then
    ShowError(Format(s_ErrSyncDirFile, [FNameSfx]), False);

  { Check equality of directory names, and rename if needed }
  Result := DoRename(DL2J.Name, DL1I.Name, FNameSfx, OpMode, BeQuieter,
                     Simulating, s_RenameDir, s_ErrRenameDir, vs_Dest)
end;


{***********************************************************************}
{ Sub-function of ProcessDirEntry, called when a file exists in both DL1
  and DL2 directory entry lists.  This function checks to see that the
  second entry IS a file (the first one is known to be, from the logic
  in ProcessDirEntry).  The algorithm used by this function is:
      1.  Compare date/time stamps of the files.
      2.  If the date/time stamps are equal, check the sizes.  If the
          sizes are not equal, copy file DL1I to DL2J.  If the sizes
          are equal, check the file attributes.  If different, copy the
          attributes from DL1I to DL2J.  Also, check to see if filenames
          are case-identical.  If not, rename DL2J to have the same case
          as DL1I.
      3.  If DL1I is newer than DL2J, copy the file DL1I to DL2J.  Check
          to see if filenames are case-identical.  If not, rename DL2J to
          be the same as DL1I.
      4.  If DL1I is older than DL2J, check the operating mode.  If OpMode
          is AbsMasterSlave, copy file DL1I to DL2J (even though DL1I is
          older), also making the filenames case-identical (DL1I to be the
          same as DL2J).  Otherwise, copy the newer file DL2J back to DL1I
          and make DL2J case-identical with DL1I.
      5.  At any time, copying a file will reset the Archive attribute on
          both files.
  This function returns True if something was actually done }
function ProcessExistingFile (const DL1I, DL2J : TDirEntry;
                              const FNameSfx : string;
                              const OpMode : TOpMode;
                              const ForceActions, BeQuieter, Simulating : boolean;
                              const vs_Dest1, vs_Dest2 : TOpModeStr) : boolean;

var
  TC, Attr : integer;

begin { ProcessExistingFile }
  Result := False;

  { Check that DL2J is a file }
  if not ((DL2J.Attr and faDirectory) = 0) then
    ShowError(Format(s_ErrSyncDirFile, [FNameSfx]), False);

  { Compare file date/time stamps }
  TC := CompareDOSTimes(DL1I.Time, DL2J.Time);
  if TC = 0 then
    begin
      { Both files have the same date/time stamp }
      if DL1I.Size <> DL2J.Size then
        begin
          { File sizes are not the same, even though date/time stamps are.
            Copy file DL1I to DL2J. }
          Result := DoCopyFile(DL1I, DL2J, FNameSfx, OpMode, ForceActions,
                               Simulating, s_CopyFile, vs_Dest2);
          Result := Result or DoRename(DL2J.Name, DL1I.Name, FNameSfx, OpMode,
                                       BeQuieter, Simulating, s_RenameFile,
                                       s_ErrRenameFile, vs_Dest2)
        end
      else
        begin
          { File sizes are the same:  set attributes if necessary }
          if DL1I.Attr <> DL2J.Attr then
            begin
              if not BeQuieter then
                write(Format(s_CopyFileAttr, [FNameSfx, vs_Dest2[OpMode]]));

              if not Simulating then
                begin
                  Attr := DL1I.Attr and not faArchive;
                  DoSetAttributes(DL1I.Name, Attr);
                  DoSetAttributes(DL2J.Name, Attr);

                  Result := True
                end;

              if not BeQuieter then
                writeln(s_Done)
            end;

          { Check filenames and rename them if necessary }
          Result := Result or DoRename(DL2J.Name, DL1I.Name, FNameSfx, OpMode,
                                       BeQuieter, Simulating, s_RenameFile,
                                       s_ErrRenameFile, vs_Dest2)
        end
    end
  else if TC > 0 then
    begin
      { DL1I has a newer date/time stamp than DL2J }
      { Copy (and rename) file DL1I to DL2J }
      Result := DoCopyFile(DL1I, DL2J, FNameSfx, OpMode, ForceActions,
                           Simulating, s_CopyFile, vs_Dest2);
      Result := Result or DoRename(DL2J.Name, DL1I.Name, FNameSfx, OpMode,
                                   BeQuieter, Simulating, s_RenameFile,
                                   s_ErrRenameFile, vs_Dest2)
    end
  else
    begin
      { DL1I has an older date/time stamp than DL2J }
      if OpMode = AbsMasterSlave then
        begin
          { Absolute Master-slave mode: force copying of older file over
            newer one }
          Result := DoCopyFile(DL1I, DL2J, FNameSfx, OpMode, ForceActions,
                               Simulating, s_CopyOlderFile, vs_Dest2);
          Result := Result or DoRename(DL2J.Name, DL1I.Name, FNameSfx, OpMode,
                                       BeQuieter, Simulating, s_RenameFile,
                                       s_ErrRenameFile, vs_Dest2)
        end
      else
        begin
          { Copy file DL2J back to DL1I, and rename if necessary }
          Result := DoCopyFile(DL2J, DL1I, FNameSfx, OpMode, ForceActions,
                               Simulating, s_CopyFileBack, vs_Dest1);
          Result := Result or DoRename(DL1I.Name, DL2J.Name, FNameSfx, OpMode,
                                       BeQuieter, Simulating, s_RenameFile,
                                       s_ErrRenameFile, vs_Dest2)
        end
    end
end;


{***********************************************************************}
{ Sub-function of ProcessDirEntry, called when a directory is to be
  removed (deleted).  This occurs only when OpMode = MasterSlave or
  AbsMasterSlave and the slave list is being processed; this routine
  assumes these conditions.  DL1I contains the directory to be removed.
  This function returns True if something was actually done }
function ProcessRemoveDir (const DL1I : TDirEntry; const FNameSfx : string;
                           const OpMode : TOpMode;
                           const ForceActions, Simulating : boolean;
                           const vs_Dest : TOpModeStr) : boolean;

var
  DLTemp : TDirEntryList;
  Count  : integer;

begin
  { NB:  This function assumes OpMode = MasterSlave or AbsMasterSlave, and
         that DL1I comes from the slave directory entry list }
  Result := False;

  write(Format(s_RemoveDir, [FNameSfx, vs_Dest[OpMode]]));

  { Check to see that the directory DL1I is truly empty.  Since the
    directory entry lists are processed in reverse order during the slave
    phase of Master-slave mode, this should be true (unless some read-only
    files were skipped) }
  DLTemp := TDirEntryList.Create;
  Count := AddDirEntries(DLTemp, DL1I.Name + '\*.*', True, AllFiles);
  DLTemp.Free;

  if Count <> 0 then
    begin
      { Directory is not empty.  This may be due to read-only files that
        were to be deleted, but were skipped because ForceActions = False }
      if (not ForceActions) or Simulating then
        begin
          writeln(s_Skipped);
          exit
        end
      else
        begin
          writeln;
          ShowError(Format(s_ErrDirNotEmpty, [DL1I.Name]), False)
        end
    end;

  { Directory is empty and may be removed }
  if not Simulating then
    begin
      if not RemoveDir(DL1I.Name) then
        begin
          writeln;
          ShowSysError(Format(s_ErrRemoveDir, [DL1I.Name]))
        end;

      Result := True
    end;

  writeln(s_Done)
end;


{***********************************************************************}
{ Sub-function of ProcessDirEntry, called when a file is to be
  removed (deleted).  This occurs only when OpMode = MasterSlave or
  AbsMasterSlave and the slave list is being processed; this routine
  assumes these conditions.  DL1I contains the file to be removed.
  This function returns True if something was actually done }
function ProcessRemoveFile (const DL1I : TDirEntry; const FNameSfx : string;
                            const OpMode : TOpMode;
                            const ForceActions, Simulating : boolean;
                            const vs_Dest : TOpModeStr) : boolean;

begin
  { NB:  This function assumes OpMode = MasterSlave or AbsMasterSlave, and
         that DL1I comes from the slave directory entry list }
  Result := False;

  write(Format(s_DeleteFile, [FNameSfx, vs_Dest[OpMode]]));

  if not Simulating then
    begin
      { Handle read-only files }
      if (DL1I.Attr and faReadOnly) <> 0 then
        if ForceActions then
          DoSetAttributes(DL1I.Name, DL1I.Attr and not faReadOnly)
        else
          begin
            { Skip copying the file }
            writeln(s_SkippedRO);
            exit
          end;

      if not DeleteFile(DL1I.Name) then
        begin
          writeln;
          ShowSysError(Format(s_ErrDeleteFile, [DL1I.Name]))
        end;

      Result := True
    end;

  writeln(s_Done)
end;


{***********************************************************************}
{ Sub-function of ProcessDirEntry, called when a new directory is to be
  created from DL1I to directory DN2.  This function returns True if
  something was actually done }
function ProcessNewDirectory (const DL1I : TDirEntry; const DN2 : string;
                              const FNameSfx : string; const OpMode : TOpMode;
                              const Simulating : boolean;
                              const vs_Dest : TOpModeStr) : boolean;

begin
  Result := False;

  write(Format(s_CreateDir, [FNameSfx, vs_Dest[OpMode]]));

  if not Simulating then
    begin
      if not CreateDir(DN2 + FNameSfx) then
        begin
          writeln;
          ShowSysError(Format(s_ErrCreateDir, [DN2 + FNameSfx]))
        end;

      Result := True
    end;

  writeln(s_Done)
end;


{***********************************************************************}
{ Sub-function of ProcessDirEntry, called when a new file is to be
  created by copying DL1I to directory DN2.  This function returns True if
  something was actually done }
function ProcessNewFile (const DL1I : TDirEntry; const DN2 : string;
                         const FNameSfx : string; const OpMode : TOpMode;
                         const ForceActions, Simulating : boolean;
                         const vs_Dest : TOpModeStr) : boolean;

var
  DL2J : TDirEntry;

begin
  { Create a suitable directory entry for the new file }
  with DL2J do
    begin
      Name := DN2 + FNameSfx;
      Attr := DL1I.Attr and not faReadOnly
    end;
  Result := DoCopyFile(DL1I, DL2J, FNameSfx, OpMode, ForceActions, Simulating,
                       s_CopyNewFile, vs_Dest)
end;


{***********************************************************************}
{ Process a single directory entry DL1[I], having directory prefix DN1.
  The algorithm for processing DL1[I] is as follows:
      1.  Find the corresponding entry in DL2:  a corresponding directory
          entry has the same filename, minus directory prefix.
          Comparisons are case-insensitive.
      2.  If this entry, DL2[J], exists, then do ONE of the following:
          a.  If the two files have the same date/time stamp, check the
              attributes.  If these are not identical, copy them across.
              Check if the filenames are case-identical.
          b.  If Absolute Master-slave mode has been selected, copy the
              first file DL1[I] across to the second directory DL2[J]
              (even if the second file is the newer one).
          c.  If the first file DL1[I] is newer, copy the file across to
              the second directory.
          d.  If the second file DL2[J] is newer, copy the file back to
              the first directory.
          After doing one of these steps, delete DL2[J] from the list (to
          prevent this routine from processing it again).
      3.  If the second entry does NOT exist, and we are NOT in Peer-peer
          mode AND we are the slave phase (IsSlave is True), delete
          the first file DL1[I].  Otherwise, copy file DL1[I] to the
          second directory DN2.
  True is returned if something was actually done. }
function ProcessDirEntry (const DN1 : string; const DL1I : TDirEntry;
                          const DN2 : string; DL2 : TDirEntryList;
                          const OpMode : TOpMode;
                          const IsSlave, ForceActions, BeQuieter,
                                Simulating : boolean;
                          const vs_Dest1, vs_Dest2 : TOpModeStr) : boolean;

var
  FNameSfx : string;
  J        : integer;
  IsDir    : boolean;

begin
  FNameSfx := TailStr(DL1I.Name, length(DN1));
  IsDir := (DL1I.Attr and faDirectory) <> 0;

  { Search for corresponding entry of DL1I in list DL2 }
  J := DirEntrySearchFN(DL2, DN2 + FNameSfx);
  if J >= 0 then
    begin
      { Entry found in both lists }
      if IsDir then
        { DL1I is a directory: process existing directories DL1I and DL2[J] }
        Result := ProcessExistingDirectory(DL1I, DL2[J], FNameSfx, OpMode,
                                           BeQuieter, Simulating, vs_Dest2)
      else
        { DL1I is a file: synchronise existing files DL1I and DL2[J] }
        Result := ProcessExistingFile(DL1I, DL2[J], FNameSfx, OpMode,
                                       ForceActions, BeQuieter, Simulating,
                                       vs_Dest1, vs_Dest2);

      { Remove entry from second list, to prevent this routine from
        processing it again }
      DL2.Delete(J)
    end
  else
    begin
      { Entry exists only in the first list }
      if (OpMode <> PeerPeer) and IsSlave then
        begin
          { Slave phase of Master-slave mode }
          if IsDir then
            { DL1I is a directory: remove directory }
            Result := ProcessRemoveDir(DL1I, FNameSfx, OpMode, ForceActions,
                                       Simulating, vs_Dest1)
          else
            { DL1I is a file: delete file }
            Result := ProcessRemoveFile(DL1I, FNameSfx, OpMode, ForceActions,
                                        Simulating, vs_Dest1)
        end
      else
        begin
          { Peer-peer mode or Master phase of Master-slave mode }
          if IsDir then
            { Create a matching directory from DL1I }
            Result := ProcessNewDirectory(DL1I, DN2, FNameSfx, OpMode,
                                          Simulating, vs_Dest2)
          else
            { Copy the new file from DL1I to DN2 }
            Result := ProcessNewFile(DL1I, DN2, FNameSfx, OpMode, ForceActions,
                                     Simulating, vs_Dest2)
        end
    end
end;


{***********************************************************************}
{ Main routine:  Synchronise directory trees DirList1 and DirList2 by
  iteratively calling ProcessDirEntry for each entry.  Display the number
  of such entries processed }
procedure Synchronise (const DirName1 : string; DirList1 : TDirEntryList;
                       const DirName2 : string; DirList2 : TDirEntryList;
                       const OpMode : TOpMode;
                       const ForceActions, BeQuieter,
                             FirstPhaseOnly, Simulating : boolean);

var
  Count, I : integer;

begin
  Count := 0;

  { Process DirList1 (master directory) first }
  writeln(Format(s_ProcessingDir, [StripBackslash(DirName1)]));
  for I := 0 to DirList1.Count - 1 do
    if ProcessDirEntry(DirName1, DirList1[I], DirName2, DirList2,
                       OpMode, False, ForceActions, BeQuieter, Simulating,
                       sa_OpModeDest1, sa_OpModeDest2) then
      inc(Count);

  { Process DirList2 (slave directory) second, if not FirstPhaseOnly }
  if not FirstPhaseOnly then
    begin
      writeln(Format(s_ProcessingDir, [StripBackslash(DirName2)]));
      if OpMode = PeerPeer then
        begin
          { For Peer-Peer mode, process in the same order as for DirList1 }
          for I := 0 to DirList2.Count - 1 do
            if ProcessDirEntry(DirName2, DirList2[I], DirName1, DirList1,
                               OpMode, False, ForceActions, BeQuieter,
                               Simulating, sa_OpModeDest2, sa_OpModeDest1) then
              inc(Count)
        end
      else
        begin
          { For the slave directory, process in reverse order so that
            deletions will work correctly }
          for I := DirList2.Count - 1 downto 0 do
            if ProcessDirEntry(DirName2, DirList2[I], DirName1, DirList1,
                               OpMode, True, ForceActions, BeQuieter,
                               Simulating, sa_OpModeDest2, sa_OpModeDest1) then
              inc(Count)
        end
    end;

  writeln(Format(s_NumEntriesProcd, [Count, BoolStr(Simulating, s_NumEntriesSim, '')]))
end;


{***********************************************************************}

var
  DirName1,
  DirName2        : string;
  OpMode          : TOpMode;
  ForceActions,
  BeQuieter,
  FirstPhaseOnly,
  Simulating      : boolean;
  DirList1,
  DirList2        : TDirEntryList;


begin { Main }
  ShowHeading;
  ProcessCmdLine(DirName1, DirName2, OpMode, ForceActions, BeQuieter,
                 FirstPhaseOnly, Simulating);

  ValidateDir(DirName1);
  ValidateDir(DirName2);
  CheckForIdenticalDirs(DirName1, DirName2);

  ShowSubHeading(DirName1, DirName2, OpMode, ForceActions, BeQuieter,
                 FirstPhaseOnly, Simulating);

  DirList1 := TDirEntryList.Create;
  DirList2 := TDirEntryList.Create;

  SearchDir(DirName1, DirList1);
  SearchDir(DirName2, DirList2);

  Synchronise(DirName1, DirList1, DirName2, DirList2, OpMode,
              ForceActions, BeQuieter, FirstPhaseOnly, Simulating);

  DirList1.Free;
  DirList2.Free;
end.
