unit DirEx;

{************************************************************************
*                                                                       *
*                       Search Directory routines                       *
*                   Copyright (C) 1997, John Zaitseff                   *
*                                                                       *
************************************************************************}

{ Author:    John Zaitseff <J.Zaitseff@unsw.edu.au>
  Date:      20th November, 1997.
  Version:   3.0
}

{ This unit provides routines to read in the contents (filenames and
  attributes) of a directory (including, if required, subdirectories).
  The resulting list may be processed by the calling application.

  This file is supplied as part of an application under the GNU General
  Public License.  Hence this file also falls under the same license.

  A directory may be thought of as consisting of a collection of
  TDirEntry objects.  Each TDirEntry object encapsulates a directory
  entry by storing the filename, attributes, size and last modified
  time.

  A collection of TDirEntry objects is managed by a TDirEntryList
  object.  This list object handles the allocation and deallocation
  of individual entries.

  Apart from the TDirEntryList object, this unit provides the
  AddDirEntries function, which searches a directory path and adds
  matching directory entries to a passed TDirEntryList object.

  The following is a typical use of the objects and function in this
  unit:

      DirEntryList := TDirEntryList.Create;
      Count := AddDirEntries(DirEntryList, 'C:\*.*', True, NormalFiles);

      for I := 0 to DirEntryList.Count - 1 do
        writeln(DirEntryList[I].Name);

      DirEntryList.Free;

}


{$IFNDEF WIN32}
Error: 'Compile under Borland Delphi for Win32 only (v2 or later).'
{$ENDIF}

{$IFNDEF OPTDEBUG}
{$UNDEF DEBUG}
{$ENDIF}

{$IFDEF DEBUG}
    {$B-,D+,I+,J-,L+,O+,P-,Q+,R+,T+,U+,V+,X+,Y+,Z1}
    {$WARNINGS ON}
    {$HINTS ON}
{$ELSE}
    {$B-,D-,I+,J-,L-,O+,P-,Q-,R-,T+,U+,V+,X+,Y-,Z1}
    {$WARNINGS ON}
    {$HINTS ON}
{$ENDIF}


interface


uses
  Classes;


type
  { Type of files to add to directory entry list }
  TDirAddType = (NormalFiles, IncludeSystemFiles, IncludeDirectories,
                 OnlyDirectories, AllFiles);

  { Individual directory entry }
  PDirEntry = ^TDirEntry;
  TDirEntry = record
                Name : string;          { Full pathname }
                Attr : integer;         { DOS attributes }
                Time : integer;         { Last modified time in DOS format }
                Size : integer;         { File size in bytes }
              end;

  { Directory entry list sort function prototype }
  TDirEntryListSortCompare = function (Item1, Item2 : PDirEntry) : integer;

  { Directory entry list object }
  TDirEntryList = class(TList)
                    protected
                      function Get (Index : integer) : TDirEntry;
                      procedure Put (Index : integer; const Item : TDirEntry);
                    public
                    { constructor Create; }
                      destructor Destroy; override;
                    { procedure Free; }

                      function Add (const Item : TDirEntry) : integer;
                      procedure Insert (Index : integer; const Item : TDirEntry);
                      procedure Delete (Index : integer);
                      procedure Clear;
                    { procedure Exchange (Index1, Index2 : integer); }
                    { procedure Move (CurIndex, NewIndex : integer); }
                      function First : TDirEntry;
                      function Last : TDirEntry;
                      procedure Sort (Compare : TDirEntryListSortCompare);

                      property Items [Index : integer] : TDirEntry
                               read Get write Put; default;
                    { property Count : integer; }
                  end;

{ TDirEntryList object description:
    Create                   - Create the directory entry list object.
    Free                     - Free the object, disposing of any entries
                               contained in it.
    Add(Item)                - Add a copy of Item onto the end of the
                               current list and return the index number of
                               where it was added.  0 is the first index.
    Insert(Index, Item)      - Insert a copy of Item onto the list just
                               BEFORE the Index'th item.
    Delete(Index)            - Delete the Index'th item, moving all the
                               other directory entries to remove the gap
                               (ie, this object does NOT model a sparse
                               list).
    Clear                    - Clear (dispose of) all of the entries.
    Exchange(Index1, Index2) - Exchange items stored at Index1 and Index2.
    Move(Index1, Index2)     - Move the Index1'th item to just before the
                               (possibly new) Index2.  This is the same as
                               calling Delete(Index1) followed by
                               Insert(Index2, Item).
    First                    - Return a copy of the first item (Items[0]).
    Last                     - Return a copy of the last item (Items[Count-1]).
    Sort(Compare)            - Sort the directory entry list based on the
                               passed Compare function.  See below for more
                               details.
    Items[Index]             - Array of directory entries (TDirEntry).  This
                               array (from 0 to Count - 1) may be read and
                               written.
    Count                    - (Read-only) Number of directory entries stored.
}


function AddDirEntries (var DirEntryList : TDirEntryList; Path : string;
                        Recurse : boolean; AddType : TDirAddType) : integer;

{ This function adds all directory entries matching Path to DirEntryList,
  which match file attributes as specified by AddAttributes.

  DirEntryList is an independently-maintained TDirEntryList object; in
  particular, it must have already been constructed.

  Path is the ambiguous (ie, containing '?' and/or '*') path to search.  A
  path consists of three parts:  the pathname, filename and extension.  The
  pathname is that part up to the last '\'; if missing, '.\' is assumed.
  The filename is that part between the pathname and just before the last '.';
  if missing, '*' is assumed.  The extension is that part from the last '.'
  onwards; if missing, '.*' is assumed.  Hence, 'C:\TEMP\' is the same as
  'C:\TEMP\*.*', while 'C:\TEMP' is the same as 'C:\TEMP.*' (the filename
  portion is present).

  Recurse specifies whether or not to recursively search subdirectories for
  matching files.  Hence, if Path='C:\Apps\Delphi\*.PAS' and Recurse=True,
  this function will add all *.PAS files found in the Delphi directory or
  any directory lower.

  AddType specifies what type of directory entries are added:
      NormalFiles         - searches for ordinary files, including those
                            that have Archive or Read-Only attributes.
      IncludeSystemFiles  - As well as those files matched by NormalFiles,
                            this also includes files that have the
                            System or Hidden attributes.
      IncludeDirectories  - As well as those files matched by
                            IncludeSystemFiles, include those files that
                            have the Directory attribute set (ie, that are
                            subdirectories).
      OnlyDirectories     - Only include files with the Directory attribute
                            set (includes all types of directories, including
                            System, Hidden and Read-Only directories)
      AllFiles            - Search for all files (ordinary, directory and
                            volume label files)

  This function returns the number of directory entries added to DirEntryList,
  or -1 if an invalid path was supplied.
}


implementation


uses
  SysUtils;


{***********************************************************************}
function AddDirEntries (var DirEntryList : TDirEntryList; Path : string;
                        Recurse : boolean; AddType : TDirAddType) : integer;

  { Internal function:  perform the actual directory searching.  Assume
    Path is a properly constructed path }
  function Add (var DirEntryList : TDirEntryList; const Path : string;
                Recurse : boolean; Attr : integer;
                AddType : TDirAddType) : integer;

  var
    Dir,
    FName     : string;
    ErrNum,
    Count,
    Count2    : integer;
    SearchRec : TSearchRec;
    DirEntry  : TDirEntry;

  begin
    Count := 0;
    Dir := ExtractFilePath(Path);
    FName := ExtractFileName(Path);

    { Main search loop for matching files }
    ErrNum := FindFirst(Path, Attr, SearchRec);
    while ErrNum = 0 do
      begin
        { Do not include '.' and '..' directory entries, or ordinary files
          if AddType = OnlyDirectories }
        if (SearchRec.Name <> '.') and (SearchRec.Name <> '..') and
           not ((AddType = OnlyDirectories) and
               ((SearchRec.Attr and faDirectory) <> faDirectory)) then
          begin
            with DirEntry do
              begin
                Name := Dir + SearchRec.Name;
                Attr := SearchRec.Attr;
                Time := SearchRec.Time;
                Size := SearchRec.Size
              end;
            DirEntryList.Add(DirEntry);
            inc(Count)
          end;
        ErrNum := FindNext(SearchRec)
      end;
    FindClose(SearchRec);

    { ErrNum 18 is ERROR_NO_MORE_FILES; ErrNum 2 is ERROR_FILE_NOT_FOUND }
    if (Count = 0) and (ErrNum <> 18) and (ErrNum <> 2) then
      begin
        Result := -1;
        exit
      end;

    { Process subdirectories, if required }
    if Recurse then
      begin
        ErrNum := FindFirst(Dir + '*.*', Attr or faDirectory, SearchRec);
        while ErrNum = 0 do
          begin
            if ((SearchRec.Attr and faDirectory) = faDirectory) and
               (SearchRec.Name <> '.') and (SearchRec.Name <> '..') then
              begin
                Count2 := Add(DirEntryList, Dir + SearchRec.Name + '\' + FName,
                              Recurse, Attr, AddType);
                if Count2 > 0 then
                  Count := Count + Count2
              end;
            ErrNum := FindNext(SearchRec)
          end;
        FindClose(SearchRec)
      end;

    Result := Count
  end;

  { Internal function:  return True if S only contains '.' characters }
  function AllDots (const S : string) : boolean;

  var
    I : integer;

  begin
    if S = '' then
      Result := False
    else
      begin
        Result := True;
        for I := 1 to length(S) do
          if S[I] <> '.' then
            begin
              Result := False;
              exit
            end
      end
  end;


const
  { AddType to DOS Attribute map }
  AddTypeAttr : array [TDirAddType] of integer =
    { NormalFiles }      ( faArchive or faReadOnly,
    { IncludeSystemFiles } faArchive or faReadOnly or faHidden or faSysFile,
    { IncludeDirectories } faArchive or faReadOnly or faHidden or faSysFile or faDirectory,
    { OnlyDirectories }    faArchive or faReadOnly or faHidden or faSysFile or faDirectory,
    { AllFiles }           faAnyFile );


var
  Dir,
  FName,
  Ext    : string;
  Attr   : integer;
  
begin { AddDirEntries }
  Dir := ExtractFilePath(Path);
  FName := ExtractFileName(Path);
  if AllDots(FName) then
    begin
      Dir := Dir + FName + '\';
      FName := '*';
      Ext := '.*'
    end
  else
    begin
      Ext := ExtractFileExt(Path);
      delete(FName, length(FName) - length(Ext) + 1, length(Ext));
      if FName = '' then
        FName := '*';
      if (Ext = '') or (Ext = '.') then
        Ext := '.*'
    end;

  Path := ExpandFileName(Dir + FName + Ext);
  if Path = '' then
    begin
      Result := -1;
      exit
    end;

  Attr := AddTypeAttr[AddType];

  Result := Add(DirEntryList, Path, Recurse, Attr, AddType);
end;


{***********************************************************************}
{ TDirEntryList object.  This object descends from TList, except that
  instead of managing a collection of pointers, it manages a set of
  directory entries.  This object manages its own allocation and
  deallocation of directory entry objects (TDirEntry).  Functions return,
  as a rule, a COPY of the directory entry instead of a pointer to it.
}

{***********************************************************************}
{ Destroy the directory entry by clearing all of the associated directory
  entries }
destructor TDirEntryList.Destroy;

begin
  Clear;
  inherited Destroy
end;

{***********************************************************************}
{ Return a copy of the Index'th entry in the list }
function TDirEntryList.Get (Index : integer) : TDirEntry;

begin
  if (Index < 0) or (Index > Count - 1) then
    Error;

  Result := PDirEntry(inherited Get(Index))^
end;

{***********************************************************************}
{ Store a copy of the passed directory entry on the list at index
  position Index }
procedure TDirEntryList.Put (Index : integer; const Item : TDirEntry);

var
  N : PDirEntry;

begin
  if (Index < 0) or (Index > Count - 1) then
    Error;

  if Assigned(List[Index]) then
    dispose(PDirEntry(List[Index]));

  new(N);
  N^ := Item;
  inherited Put(Index, Pointer(N))
end;

{***********************************************************************}
{ Add a directory entry to the end of the list and return the index
  position.  Note that the first entry has an index of 0. }
function TDirEntryList.Add (const Item : TDirEntry) : integer;

var
  N : PDirEntry;

begin
  new(N);
  N^ := Item;
  Result := inherited Add(Pointer(N))
end;

{***********************************************************************}
{ Insert a copy of the passed directory entry into the list, just
  BEFORE the Index'th item }
procedure TDirEntryList.Insert (Index : integer; const Item : TDirEntry);

var
  N : PDirEntry;

begin
  if (Index < 0) or (Index > Count) then
    Error;

  new(N);
  N^ := Item;
  inherited Insert(Index, Pointer(N))
end;

{***********************************************************************}
{ Delete the directory entry stored at Index in the list.  Move all of
  the following entries down one place }
procedure TDirEntryList.Delete (Index : integer);

begin
  if (Index < 0) or (Index >= Count) then
    Error;

  if Assigned(List[Index]) then
    dispose(PDirEntry(List[Index]));

  inherited Delete(Index)
end;

{***********************************************************************}
{ Clear all of the directory entries currently contained in the list }
procedure TDirEntryList.Clear;

var
  I : integer;

begin
  for I := 0 to Count - 1 do
    if Assigned(List[I]) then
      dispose(PDirEntry(List[I]));

  inherited Clear
end;

{***********************************************************************}
{ Return a copy of the first directory entry stored in the list (ie,
  the one at index 0) }
function TDirEntryList.First : TDirEntry;

begin
  Result := Get(0)
end;

{***********************************************************************}
{ Return a copy of the last directory entry stored in the list (ie,
  the one at index Count-1) }
function TDirEntryList.Last : TDirEntry;

begin
  Result := Get(Count - 1)
end;

{***********************************************************************}
{ Sort the collection of directory entries based on the Compare
  function pointer.  Compare(P1, P2) takes two pointers to directory
  entries and returns a value based on their relative positions:
  if P1 < P2, Compare returns a value less than 0.  Similarly, Compare
  returns >0 if P1 > P2, and 0 if P1 = P2.  A typical Compare function
  may compare P1^.Filename and P2^.Filename }
procedure TDirEntryList.Sort (Compare : TDirEntryListSortCompare);

begin
  inherited Sort(TListSortCompare(Compare))
end;


end.

