unit ParamLn;

{************************************************************************
*                                                                       *
*                        Parameter Line Routines                        *
*                   Copyright (C) 1997, John Zaitseff                   *
*                                                                       *
************************************************************************}

{ Author:    John Zaitseff <J.Zaitseff@unsw.edu.au>
  Date:      10th April, 1998.
  Version:   2.2

  The previous version of these routines was written in 1994 for use under
  Borland Pascal v7.  Version 2.0 (1997) is the first rewrite of these for
  Win32 programs.

  This file is supplied as part of an application under the GNU General
  Public License.  Hence this file also falls under the same license.

  This unit provides a number of routines to deal with the parameter line
  passed to any Win32 program.  The routines are considerably more power-
  ful than using ParamStr and ParamCount.

  The parameter line may be thought of as a series of words separated by
  separators.  The separators are spaces, tabs, commas, semicolons, slashes
  ("/"), plus signs and equals signs.  Note that having multiple separators
  (except for spaces or tabs) implies null words between them.  Words within
  double quotes are treated as a single word; these quotes are NOT treated
  as separators.

  Par.Line contains the full parameter line as a string.  Note that the
      first word that appears is treated as the program name.
  Par.Count returns the number of words on the parameter line.  Some of these
      words may be null words, as described above.  This does not count
      the program name (the 0'th word).
  Par.Str(N) returns the N'th word from the command line.  If the word
      starts with the "/" separator, the separator is included as part of
      the word.  Note that word 0 is the program name.
  Par.Sep(N) returns the separator just BEFORE the N'th word.
  Par.Option(Opt, N) returns the index of the first word containing the
      specified option (ie, a word starting with "/"), starting after the
      N'th word.  0 is returned if no such option exists.  Note that a
      case-insensitive search is done.  Also note that Opt should not
      have the leading option character (it is internally ignored if present).
  Par.IsOption(N) returns True if the N'th word is an option.
  Par.IsTheOption(N, Opt) returns True if the N'th word is the same as Opt.
      Note that a case-insensitive test is done.  Also note that Opt should
      not have the leading option character (it is internally ignored if
      present).

  For example:

      If Par.Line = 'C:\TEST.EXE *.pas /h/X,,nul /o=C/h="Test file.lst"',

          Par.Count = 9
          Par.Str(0) = 'C:\TEST.EXE'
          Par.Str(1) = '*.pas'           Par.Sep(1) = ' '
          Par.Str(2) = '/h'              Par.Sep(2) = '/'
          Par.Str(3) = '/X'              Par.Sep(3) = '/'
          Par.Str(4) = ''                Par.Sep(4) = ','
          Par.Str(5) = 'nul'             Par.Sep(5) = ','
          Par.Str(6) = '/o'              Par.Sep(6) = '/'
          Par.Str(7) = 'C'               Par.Sep(7) = '='
          Par.Str(8) = '/h'              Par.Sep(8) = '/'
          Par.Str(9) = 'Test file.lst'   Par.Sep(9) = '='
                                         Par.Sep(10) = ' '
          Par.Option('h', 0)  = 2
          Par.Option('H', 0)  = 2        Par.IsOption(1) = False
          Par.Option('/H', 0) = 2        Par.IsOption(2) = True
          Par.Option('/h', 2) = 8        Par.IsTheOption(1, 'H') = False
          Par.Option('?', 0)  = 0        Par.IsTheOption(2, 'H') = True
}


{$IFNDEF WIN32}
Error: 'Compile under Borland Delphi for Win32 only (v2 or later).'
{$ENDIF}

{$IFNDEF OPTDEBUG}
{$UNDEF DEBUG}
{$ENDIF}

{$IFDEF DEBUG}
    {$B-,D+,I+,J-,L+,O+,P-,Q+,R+,T+,U+,V+,X+,Y+,Z1}
    {$WARNINGS ON}
    {$HINTS ON}
{$ELSE}
    {$B-,D-,I+,J-,L-,O+,P-,Q-,R-,T+,U+,V+,X+,Y-,Z1}
    {$WARNINGS ON}
    {$HINTS ON}
{$ENDIF}


interface


type
  TCharSet = set of char;

const
  DefOptionChars      = ['/'];
  DefNonSpaceSepChars = [',', ';', '+', '='];
  DefSpaceSepChars    = [' ', ^I, ^J, ^M];
  DefQuoteChars       = ['"'];

type
  { Internal parsed representation definition }
  PParsed  = ^TParsed;
  TParsed  = record
               Index : integer;
               Str   : string;
               Sep   : char;
               Next  : PParsed;
             end;

  { Parameter line object definition }
  TPar     = class
               private
                 FParLine     : string;
                 FCount       : integer;
                 FParsed      : PParsed;
                 FOptionChars : TCharSet;

               public
                 constructor Create   (const ParLine           : string;
                                       const OptionChars,
                                             NonSpaceSepChars,
                                             SpaceSepChars,
                                             QuoteChars        : TCharSet);
                 destructor Destroy; override;

                 property Line : string read FParLine;
                 property Count : integer read FCount;

                 function Str (N : integer) : string;
                 function Sep (N : integer) : char;
                 function Option (Opt : string; Index : integer) : integer;
                 function IsOption (N : integer) : boolean;
                 function IsTheOption (N : integer; Opt : string) : boolean;
             end;

var
  Par : TPar;         { Parameter line object }


implementation

uses
  SysUtils;

{************************************************************************}
{ Create the TPar object.  ParLine is the initial parameter line string.
  The character sets specify, respectively, what characters are considered
  to be the start of options, part of non-space separators, part of white
  space and part of quotes.  Option characters are treated as separators,
  but are included as part of parameter words.  Null words may occur
  between multiple occurrences of non-space separators.  Multiple white
  space separators are treated as a single separator. }
constructor TPar.Create (const ParLine : string;
                         const OptionChars, NonSpaceSepChars,
                               SpaceSepChars, QuoteChars : TCharSet);

var
  I, L     : integer;
  S        : string;
  C, SepC,
  PrevSepC : char;
  P, PPrev : PParsed;

begin
  inherited Create;

  FParLine := ParLine;
  FCount := 0;
  FParsed := nil;
  FOptionChars := OptionChars;

  PPrev := nil;

  I := 1;
  L := length(FParLine);
  SepC := ' ';

  { Skip any leading spaces }
  while (I <= L) and (FParLine[I] in SpaceSepChars) do
    inc(I);

  { Iterate over the entire string }
  while I <= L do
    begin
      PrevSepC := SepC;

      { Make an option character part of the returned word }
      if FParLine[I] in OptionChars then
        begin
          S := FParLine[I];
          PrevSepC := FParLine[I];
          inc(I)
        end
      else
        S := '';

      { Iterate over an entire word }
      while I <= L do
        begin
          C := FParLine[I];

          if C in QuoteChars then
            begin
              { Copy characters until matching quote found, or end of
                string reached.  Do not include actual quote characters }
              inc(I);
              while (I <= L) and (FParLine[I] <> C) do
                begin
                  S := S + FParLine[I];
                  inc(I)
                end;
              if (I <= L) and (FParLine[I] = C) then
                inc(I)
            end
          else if C in SpaceSepChars then
            begin
              { Skip any spaces after a word }
              while (I <= L) and (FParLine[I] in SpaceSepChars) do
                inc(I);
              SepC := ' ';
              break { while loop }
            end
          else if C in NonSpaceSepChars then
            begin
              { Non-space separator found: the start of a new word }
              SepC := C;
              inc(I);
              break { while loop }
            end
          else if C in OptionChars then
            begin
              { Separator found (option char); the start of a new word.
                Do not skip this character as it is to be included as
                part of the word }
              SepC := C;
              break { while loop }
            end
          else
            begin
              { Ordinary character.  At it to the word being scanned }
              S := S + C;
              inc(I)
            end
        end; { while loop for word }

      { Add the word, if it has been found, to our list }
      if (S <> '') or (I <= L + 1) then
        begin
          new(P);
          P^.Index := FCount;
          P^.Str := S;
          P^.Sep := PrevSepC;
          P^.Next := nil;
          inc(FCount);

          if PPrev = nil then
            FParsed := P
          else
            PPrev^.Next := P;
          PPrev := P
        end
    end; { while loop for entire string }

  { Check if last character is a non-space separator.  If so, add a null
    word to our list }
  if FParLine[L] in NonSpaceSepChars then
    begin
      new(P);
      P^.Index := FCount;
      P^.Str := '';
      P^.Sep := FParLine[L];
      P^.Next := nil;
      inc(FCount);

      if PPrev = nil then
        FParsed := P
      else
        PPrev^.Next := P
    end;

  { FCount does NOT include word 0 (the program name) in its count }
  dec(FCount)
end;

{************************************************************************}
{ Destroy the TPar object.  Namely, clean up the associated list of words }
destructor TPar.Destroy;

var
  P, PNext : PParsed;

begin
  P := FParsed;
  while P <> nil do
    begin
      PNext := P^.Next;
      dispose(P);
      P := PNext
    end;
  FParsed := nil;

  inherited Destroy
end;

{************************************************************************}
{ Return the N'th word (parameter) in the command line.  Note that the
  option separator is included as part of any word.  The 0'th word is
  the program name }
function TPar.Str (N : integer) : string;

var
  P : PParsed;

begin
  Result := '';
  if (N >= 0) and (N <= FCount) then
    begin
      P := FParsed;
      while P <> nil do
        if P^.Index = N then
          begin
            Result := P^.Str;
            exit
          end
        else
          P := P^.Next
    end
end;

{************************************************************************}
{ Return the separator just BEFORE the N'th word }
function TPar.Sep (N : integer) : char;

var
  P : PParsed;

begin
  Result := ' ';
  if (N >= 0) and (N <= FCount) then
    begin
      P := FParsed;
      while P <> nil do
        if P^.Index = N then
          begin
            Result := P^.Sep;
            exit
          end
        else
          P := P^.Next
    end
end;

{************************************************************************}
{ Return the index of the first word containing the specified option
  (ie, a word starting with an option character), starting AFTER the
  N'th word.  0 is returned if no such option exists.  Note that a
  case-insensitive search is done.  If Opt has a leading option character,
  it is ignored }
function TPar.Option (Opt : string; Index : integer) : integer;

var
  P : PParsed;

begin
  Result := 0;

  if Index < 0 then
    Index := 0
  else if Index >= FCount then
    exit;

  if (length(Opt) > 0) and (Opt[1] in FOptionChars) then
    delete(Opt, 1, 1);

  P := FParsed;
  while P <> nil do
    if (P^.Index > Index) and (P^.Sep in FOptionChars) and
       (CompareText(Opt, copy(P^.Str, 2, length(P^.Str)-1)) = 0) then
      begin
        Result := P^.Index;
        exit
      end
    else
      P := P^.Next
end;

{************************************************************************}
{ Return True if the N'th word is an option (ie, starts with an option
  character }
function TPar.IsOption (N : integer) : boolean;

var
  S : string;

begin
  S := Self.Str(N);
  if S <> '' then
    Result := (S[1] in FOptionChars)
  else
    Result := False
end;

{************************************************************************}
{ Return True if the N'th word is an option matching Opt.  Note that a
  case-insensitive comparison is made.  If Opt has a leading option
  character, it is ignored }
function TPar.IsTheOption (N : integer; Opt : string) : boolean;

var
  S : string;

begin
  Result := False;

  S := Self.Str(N);
  if (S <> '') and (S[1] in FOptionChars) then
    begin
      if (length(Opt) > 0) and (Opt[1] in FOptionChars) then
        delete(Opt, 1, 1);

      Result := (CompareText(Opt, copy(S, 2, length(S)-1)) = 0)
    end
end;


{************************************************************************}

initialization
  Par := TPar.Create(CmdLine, DefOptionChars, DefNonSpaceSepChars,
                     DefSpaceSepChars, DefQuoteChars)
finalization
  Par.Free
end.
